#include <stdio.h>
#include <sys/socket.h>

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"

#include "wifi.h"
#include "camera.h"
//#include "webserver.h"
#include "esp_timer.h"
#include "esp_camera.h"

#include "nvs_flash.h"

#include "protocol.h"

/* void send_frame(int client_fd, size_t len_buffer) { */
/*   //printf("send NEW_FILE: %zu\n", len_buffer); */
/*   int64_t times[6]; */
/*   int64_t start, end; */
  
/*   start = esp_timer_get_time(); */
/*   int64_t start_frame = start; */
/*   times[0] = esp_timer_get_time() - start; */

/*   start = esp_timer_get_time(); */
/*   if (!fb) { */
/*     fatal("Camera capture failed"); */
/*   } */
/*   times[1] = esp_timer_get_time() - start; */

/*   start = esp_timer_get_time(); */
/*   send_data(client_fd, fb->buf, len_buffer); */
/*   times[2] = esp_timer_get_time() - start; */

/*   start = esp_timer_get_time(); */
/*   if (fb) { */
/*     esp_camera_fb_return(fb); */
/*   } */
/*   times[3] = esp_timer_get_time() - start; */
/*   /\* int64_t fr_end = esp_timer_get_time(); *\/ */
/*   /\* uint32_t frame_time = (uint32_t) fr_end - fr_start; *\/ */

/*   start = esp_timer_get_time(); */
/*   send_ctrl(client_fd, FILE_DONE, 0); */
/*   int64_t end_frame = esp_timer_get_time(); */
/*   times[4] = end_frame - start; */
/*   times[5] = 33 - (end_frame - start_frame)/1000; */

/*   send_data(client_fd, (unsigned char *) times, sizeof(times)); */

/*   /\* if (time > 0) { *\/ */
/*   /\*   vTaskDelay(pdMS_TO_TICKS(33 - time)); *\/ */
/*   /\* } *\/ */
/* } */

void start_camera_client(int client_fd) {
  //  int64_t start = esp_timer_get_time();

  while (1) {
    uint32_t nr_frames = expect(client_fd, FRAMES);
    /* printf("got a frames message with %ld frames\n", nr_frames); */

    for (int i = 0; i < nr_frames; i++) {

      int64_t start = esp_timer_get_time();

      /* puts("getting a frame"); */
      camera_fb_t *fb = esp_camera_fb_get();
      if (!fb) {
	printf("Camera capture failed");
	return;
      }

      send_ctrl(client_fd, NEW_FRAME, (uint32_t) fb->len);
      /* printf("just sent a new frame msg with %zu\n", fb->len); */

      send_data(client_fd, fb->buf, (uint32_t) fb->len);
      /* puts("sent data"); */

      if (fb) {
	esp_camera_fb_return(fb);
      }

      int64_t end = esp_timer_get_time();

      int duration = (int) ((end - start)/ 1000);
      int time_to_wait = FRAME_MS - duration;
      if (time_to_wait > 0) {
	vTaskDelay(pdMS_TO_TICKS(time_to_wait));
	/* printf("waited %d\n", time_to_wait); */
      }
    }
  }
}

int create_socket() {
  int client_fd = socket(AF_INET, SOCK_STREAM, 0);
  if (client_fd == -1) {
    fatal("Failed to create client socket\n");
  }

  // Connect to the server
  struct sockaddr_in server_addr = {0};
  server_addr.sin_family = AF_INET;
  server_addr.sin_port = htons(12345);
  server_addr.sin_addr.s_addr = inet_addr("192.168.178.97");
  if (connect(client_fd, (struct sockaddr *) &server_addr, sizeof(server_addr)) == -1) {
    fatal("Failed to connect to server");
  }

  int buffer_size = 1024 * 1024; // 1 MB buffer size
  setsockopt(client_fd, SOL_SOCKET, SO_SNDBUF, &buffer_size, sizeof(buffer_size));

  return client_fd;
}

void app_main() {
    //Initialize NVS
    esp_err_t ret = nvs_flash_init();
    if (ret == ESP_ERR_NVS_NO_FREE_PAGES || ret == ESP_ERR_NVS_NEW_VERSION_FOUND) {
      ESP_ERROR_CHECK(nvs_flash_erase());
      ret = nvs_flash_init();
    }
    ESP_ERROR_CHECK(ret);

    ESP_ERROR_CHECK(wifi_init());

    ESP_ERROR_CHECK(setup_camera());

    int client_fd = create_socket();
    
    start_camera_client(client_fd);

    close(client_fd);

    while (true) {
      vTaskDelay(1);
    }
}

