/*
  ESP32 I2S Microphone Sample
  esp32-i2s-mic-sample.ino
  Sample sound from I2S microphone, display on Serial Plotter
  Requires INMP441 I2S microphone
 
  DroneBot Workshop 2022
  https://dronebotworkshop.com
*/
 
// Include I2S driver
#include <driver/i2s.h>

// select is GPIO12
// ws is GPIO13
// data is GPIO15
// BCLK Is GPIO14

// Connections to INMP441 I2S microphone
#define I2S_WS 13
#define I2S_SD 15
#define I2S_SCK 14
#define I2S_SEL 12
 
// Use I2S Processor 0
#define I2S_PORT I2S_NUM_0
 
// Define input buffer length
#define bufferLen 64
int32_t sBuffer[bufferLen];

void check(esp_err_t err) {
	if (err != ESP_OK) {
		Serial.print("error: ");
		Serial.println(esp_err_to_name(err));
	}
	else {
		Serial.println("ok");
	}
}
 
void i2s_install() {
  // Set up I2S Processor configuration
  const i2s_config_t i2s_config = {
    .mode = i2s_mode_t(I2S_MODE_MASTER | I2S_MODE_RX),
    .sample_rate = 44100,
    .bits_per_sample = i2s_bits_per_sample_t(32),
    .channel_format = I2S_CHANNEL_FMT_RIGHT_LEFT,
    .communication_format = i2s_comm_format_t(I2S_COMM_FORMAT_STAND_I2S),
    .intr_alloc_flags = 0,
    .dma_buf_count = 8,
    .dma_buf_len = bufferLen,
    .use_apll = false
  };
 
	check(i2s_driver_install(I2S_PORT, &i2s_config, 0, NULL));
}
 
void i2s_setpin() {
  // Set I2S pin configuration
  const i2s_pin_config_t pin_config = {
    .bck_io_num = I2S_SCK,
    .ws_io_num = I2S_WS,
    .data_out_num = -1,
    .data_in_num = I2S_SD
  };
 
  check(i2s_set_pin(I2S_PORT, &pin_config));
}
 
void setup() {
 
  // Set up Serial Monitor
  Serial.begin(115200);

	delay(10000);
	
  Serial.println("Starting serial");

	
 
  delay(1000);
 
  // Set up I2S
  i2s_install();
  i2s_setpin();
  check(i2s_start(I2S_PORT));
 
 
  delay(500);

	digitalWrite(I2S_SEL, 0);
}
 
void loop() {
 
  // False print statements to "lock range" on serial plotter display
  // Change rangelimit value to adjust "sensitivity"
  int rangelimit = 3000;
  /* Serial.print(rangelimit * -1); */
  /* Serial.print(" "); */
  /* Serial.print(rangelimit); */
  /* Serial.print(" "); */
 
  // Get I2S data and place in data buffer
  size_t bytesIn = 0;
  esp_err_t result = i2s_read(I2S_PORT, &sBuffer, bufferLen * sizeof(int32_t), &bytesIn, portMAX_DELAY);
	//check(result);
 
  if (result == ESP_OK)
  {
		/* Serial.print("bufferLen: "); */
		/* Serial.print(bufferLen); */

		/* Serial.print("bytesIn: "); */
		/* Serial.print(bytesIn); */
		/* Serial.print(" "); */
		
    // Read I2S data buffer
    int16_t samples_read = bytesIn / 8;
    if (samples_read > 0) {
      float mean = 0;
      for (int16_t i = 0; i < samples_read; ++i) {
        mean += (sBuffer[i]);
				//Serial.println(sBuffer[i]);
      }
 
      // Average the data reading
      mean /= samples_read;
 
      // Print to serial plotter
			/* Serial.print(samples_read); */
			/* Serial.print(" "); */
			/* Serial.print(sBuffer[0]); */
			/* Serial.print(" "); */
      Serial.println(mean);
    }
  }
}
