#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>

#include <time.h>

#include "protocol.h"

#define MAX_LENGTH_FILENAME 100


unsigned char *read_file(char *filename, uint32_t *len_buffer) {
    FILE *file = fopen(filename, "rb");
    if (file == NULL) {
      fatal("Failed to open input file");
    }

    // Get the size of the input file
    fseek(file, 0, SEEK_END);
    long filesize = ftell(file);
    fseek(file, 0, SEEK_SET);

    // Allocate a buffer for the file data
    unsigned char *buffer = (unsigned char *) malloc(filesize);
    if (buffer == NULL) {
      fatal("Failed to allocate memory for file data");
    }
    *len_buffer = (uint32_t) filesize;

    // Read the file data into the buffer
    size_t bytes_read = fread(buffer, 1, filesize, file);
    if (bytes_read != filesize) {
      fatal("Failed to read file data");
    }
    if (ferror(file)) {
      fatal("Failed to read file");
    }

    // Close the input file
    fclose(file);

    return buffer;
}


/* int send_file(int client_fd, char *filename) { */


/*   send_ctrl(client_fd, FILE_DONE, 0); */
  
/*   puts("done"); */

/*   return 0; */
/* } */


int main() {
  // Create the client socket
  int client_fd = socket(AF_INET, SOCK_STREAM, 0);
  if (client_fd == -1) {
    fprintf(stderr, "Error: Failed to create client socket\n");
    return 1;
  }

  // Connect to the server
  struct sockaddr_in server_addr = {0};
  server_addr.sin_family = AF_INET;
  server_addr.sin_port = htons(12345);
  server_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
  if (connect(client_fd, (struct sockaddr *) &server_addr, sizeof(server_addr)) == -1) {
    fprintf(stderr, "Error: Failed to connect to server\n");
    return 1;
  }

  int frame = 0;
  char filename[MAX_LENGTH_FILENAME];

  while (1) {
    uint32_t nr_frames = expect(client_fd, FRAMES);
    printf("got a frames message with %d frames\n", nr_frames);

    for (int i = 0; i < nr_frames; i++) {
      clock_t start_time, end_time;
      double elapsed_time_ms;

      // Record start time
      start_time = clock();

      snprintf(filename, MAX_LENGTH_FILENAME, "input/output_%04d.jpg", frame + 1);
      frame++;
      if (frame == 96) frame = 0;

      uint32_t len_buffer;
      puts("getting a frame");
      unsigned char *buffer = read_file(filename, &len_buffer);
      //printf("read a file into buffer, length: %d\n", len_buffer);

      send_ctrl(client_fd, NEW_FRAME, len_buffer);
      printf("just sent a new frame msg with %zu\n", len_buffer);

      send_data(client_fd, buffer, len_buffer);
      puts("sent data");

      free(buffer);

      // Record end time
      end_time = clock();

      // Calculate elapsed time in ms
      elapsed_time_ms = (double) (end_time - start_time) * 1000 / CLOCKS_PER_SEC;

      int time_to_wait = (int) (FRAME_MS - elapsed_time_ms);
      
      if (time_to_wait > 0) {
	usleep(time_to_wait * 1000);
	printf("waited %d\n", time_to_wait);
      }
      

    }
  }

  close(client_fd);
  
  return 0;
}


