#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <stdarg.h>

#include <assert.h>

#include "protocol.h"
#include "device.h"

#if ESP_RUNNING
#include "esp_system.h"
#endif

void fatal(const char *fmt, ...) {
    va_list args;
    char buf[1024];  // Buffer to hold the formatted message
    va_start(args, fmt);
    vsnprintf(buf, sizeof(buf), fmt, args);  // Format the message into the buffer
    va_end(args);
    perror(buf);  // Pass the buffer as the argument to perror
#if ESP_RUNNING
    esp_restart();
#else
    exit(EXIT_FAILURE);
#endif
}

void send_ctrl(int fd, uint32_t type, uint32_t data) {
  ctrl_msg_t msg;
  msg.type = type;
  msg.data = data;
  ssize_t bytes_sent = send(fd, &msg, sizeof(msg), 0);
  if (bytes_sent == -1) {
    fatal("Failed to send data to server");
  }
}

void send_data(int fd, unsigned char *buffer, size_t len) {
  size_t total_bytes_sent = 0;
  while (total_bytes_sent < len) {
    size_t bytes_to_sent = LEN_DATA;
    if (len - total_bytes_sent < LEN_DATA) {
      bytes_to_sent = len - total_bytes_sent;
    }
    ssize_t bytes_sent = send(fd, buffer + total_bytes_sent, bytes_to_sent, 0);
    if (bytes_sent == -1) {
      fatal("Failed to send data to server");
    }
    total_bytes_sent += bytes_sent;
  }
}

void receive_data(int fd, unsigned char *buffer, size_t len) {
  size_t total_bytes_received = 0;
  while (total_bytes_received < len) {
    size_t bytes_to_recv = LEN_DATA;
    if (len - total_bytes_received < LEN_DATA) {
      bytes_to_recv = len - total_bytes_received;
    }
    ssize_t bytes_received = recv(fd, buffer + total_bytes_received, bytes_to_recv, 0);
    //printf("bytes_received %zd\n", bytes_received);
    if (bytes_received == -1) {
      fatal("Failed to send data to server");
    }
    else if (bytes_received == 0) {
      fatal("Connection closed");
    }
    else {
      total_bytes_received += bytes_received;
    }
  }
}

uint32_t receive_ctrl(int fd, ctrl_msg_t *msg) {
  ssize_t bytes_received = recv(fd, msg, sizeof(ctrl_msg_t), MSG_WAITALL);
  printf("bytes_received %zd\n", bytes_received);
  if (bytes_received == -1) {
    fatal("Failed to send data to server");
  }
  else if (bytes_received == 0) {
    return CLOSE_CONNECTION;
  }
  
  return msg->type;
}

uint32_t expect(int fd, uint32_t type) {
  ctrl_msg_t msg;
  if (receive_ctrl(fd, &msg) != type) {
    fatal("Expected message of type %u instead of ", type, msg.type);
  }
  return msg.data;
}

/*
unsigned char *alloc_buffer(size_t len_buffer) {
  size_t len = (len_buffer + (LEN_DATA - 1)) / LEN_DATA * LEN_DATA;
  printf("allocating a buffer with size %zu\n", len_buffer);
  printf("becomes size %zu\n", len);
  return (unsigned char *) malloc(len);
}
*/
