#include <microhttpd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <pthread.h>
#include <signal.h>

#include "protocol.h"
#include "stream-server.h"
#include "mux.h"

#define PAGE_TEMPLATE "<html><body><img src=\"/stream\"></body></html>"

const char *page = "<!DOCTYPE html><html><head>"
  "<title>Camera Director</title>"
  "<style>"
  "  /* CSS styles */"
  "  body {"
  "    display: flex;"
  "    flex-direction: column;"
  "    align-items: center;"
  "    justify-content: center;"
  "    height: 100vh;"
  "    margin: 0;"
  "    background-color: #f9f9f9; /* Light gray background color */"
  "  }"
  "  #buttons {"
  "    display: flex;"
  "    justify-content: center;"
  "    margin-bottom: 20px;"
  "  }"
  "  button {"
  "    padding: 10px 20px;"
  "    background-color: #2196f3; /* Blueish background color */"
  "    color: #fff; /* White text color */"
  "    border: none; /* Remove button borders */"
  "    border-radius: 5px; /* Rounded corners */"
  "    box-shadow: 2px 2px 5px rgba(0, 0, 0, 0.2); /* Box shadow effect */"
  "    transition: background-color 0.3s ease; /* Smooth background color transition */"
  "    margin-right: 10px; /* Add space between buttons */"
  "    width: 120px; /* Set button width */"
  "    text-align: center; /* Center button text */"
  "  }"
  "  button:hover {"
  "    background-color: #1976d2; /* Darker blueish background color on hover */"
  "  }"
  "button:active {"
  "  background-color: #0d47a1; /* Even darker blueish background color when pressed */"
  "  box-shadow: none; /* Remove box shadow when pressed */"
  "  transform: translateY(2px); /* Push the button down slightly when pressed */"
  "}  "
  "h1 {"
  "  font-size: 48px;"
  "  font-family: 'Roboto', sans-serif; /* Use Roboto font */"
  "  color: #333333; /* Dark gray color */"
  "  margin-top: 20px; /* Add spacing above the title */"
  "}  "
  "  #container {"
  "    display: flex;"
  "    flex-direction: column;"
  "    align-items: center;"
  "    min-height: 100vh; /* Set the minimum height to 100% viewport height */"
  "  }  "
  "  #content {"
  "    display: flex;"
  "    justify-content: center;"
  "    width: 80%; /* Set the width of the column */"
  "    padding: 20px;"
  "    background-color: #f1f1f1; /* Light gray background color */"
  "    border-radius: 10px; /* Rounded corners */"
  "    box-shadow: 2px 2px 5px rgba(0, 0, 0, 0.2);"
  "    height: 100%;"
  "  }"
  "</style>"
  "</head><body>"
  "<div id=\"container\">"
  "<h1 id=\"banner\">"
  "Direct your own movie!"
  "</h1>"
  "<div id=\"buttons\">"
  "<button id=\"cam1\">Camera 1</button>"
  "<button id=\"cam2\">Camera 2</button>"
  "<button id=\"stop\">Create</button>"
  "</div>"
  "<div id=\"content\"></div>"
  "</div>"
  "<script>"
  "document.getElementById('cam1').addEventListener('click', function() {"
  "  var xhr = new XMLHttpRequest();"
  "  xhr.open('GET', '/cam1');"
  "  xhr.send();"
  "});"
  "document.getElementById('cam2').addEventListener('click', function() {"
  "  var xhr = new XMLHttpRequest();"
  "  xhr.open('GET', '/cam2');"
  "  xhr.send();"
  "});"
  "document.getElementById('stop').addEventListener('click', function() {"
  "  var xhr = new XMLHttpRequest();"
  "  xhr.onload = function() {"
  "    document.getElementById('content').innerHTML = xhr.responseText;"
  "  };"
  "  xhr.open('GET', '/stop');"
  "  xhr.send();"
  "});"
  "</script></body></html>";

#define MAX_LENGTH_FILENAME 100



unsigned char *read_file(char *filename, size_t *len_buffer) {
  FILE *file = fopen(filename, "rb");
  if (file == NULL) {
    fatal("Failed to open input file");
  }

  // Get the size of the input file
  fseek(file, 0, SEEK_END);
  long filesize = ftell(file);
  fseek(file, 0, SEEK_SET);

  // Allocate a buffer for the file data
  unsigned char *buffer = (unsigned char *) malloc(filesize);
  if (buffer == NULL) {
    fatal("Failed to allocate memory for file data");
  }
  *len_buffer = filesize;

  // Read the file data into the buffer
  size_t bytes_read = fread(buffer, 1, filesize, file);
  if (bytes_read != filesize) {
    fatal("Failed to read file data");
  }
  if (ferror(file)) {
    fatal("Failed to read file");
  }

  // Close the input file
  fclose(file);

  return buffer;
}



#define SZ_RESPONSE 4000


void create_video_part(int nr_vids, char *response) {
    int nr_chars = snprintf(response, SZ_RESPONSE, "<div>\n");
    for (int i = 0; i < nr_vids; i++) {
      nr_chars += snprintf(response + nr_chars, SZ_RESPONSE - nr_chars, "<video controls><source src=\"vid%d.mp4\" type=\"video/mp4\"></video>", i);
    }
    nr_chars += snprintf(response + nr_chars, SZ_RESPONSE - nr_chars, "</div>\n");
}

int empty_response(struct MHD_Connection *connection) {
  struct MHD_Response *response = MHD_create_response_from_buffer(0, NULL, MHD_RESPMEM_PERSISTENT);
  MHD_add_response_header(response, "Content-Type", "text/plain");
  int ret = MHD_queue_response(connection, MHD_HTTP_OK, response);
  MHD_destroy_response(response);
  return ret;
}




/* pthread_t threads[NR_CAMS]; */
/* volatile int streaming[NR_CAMS]; */

/* int is_thread_running(pthread_t thread) { */
/*   return pthread_kill(thread, 0) == 0; */
/* } */


pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t cond = PTHREAD_COND_INITIALIZER;
volatile int running[NR_CAMS] = {0, 0};


void write_binary_buffer(unsigned char* buffer, size_t len_buffer, char* filename) {
    FILE* file = fopen(filename, "wb"); // Open the file in binary write mode
    if (file == NULL) {
        perror("Error opening file");
        return;
    }
    size_t bytes_written = fwrite(buffer, sizeof(unsigned char), len_buffer, file);
    if (bytes_written != len_buffer) {
      fatal("Error writing to file");
    }
    fclose(file);
}

volatile int someone_busy = 0;

void get_frames(int i, int nr_frames) {
  static int frame = 0;
  int cam_fd = cam_fds[i];
  send_ctrl(cam_fd, FRAMES, nr_frames);

  char filename[MAX_LENGTH_FILENAME];

  while (someone_busy);
  someone_busy = 1;

  for (int i = 0; i < nr_frames; i++) {
    uint32_t len_buffer = expect(cam_fd, NEW_FRAME);
    printf("got a NEW_FRAME with length: %d\n", len_buffer);
    unsigned char *buffer = (unsigned char *) malloc(len_buffer);
    receive_data(cam_fd, buffer, len_buffer);
    //snprintf(filename, MAX_LENGTH_FILENAME, "output_%04d.jpg", frame + 1);
    //frame++;
    //write_binary_buffer(buffer, len_buffer, filename);

    video_st.jpg_packet->data = buffer;
    video_st.jpg_packet->size = len_buffer;
    write_video_frame(oc, &video_st);
    
    free(buffer);
  }
  someone_busy = 0;
}

void *handle_cam(void *id) {
  int i = *(int *) id;

  while (1) {
    pthread_mutex_lock(&mutex);
    while (!running[i]) {
      pthread_cond_wait(&cond, &mutex);
    }
    pthread_mutex_unlock(&mutex);
    get_frames(i, 10);
  }
  return NULL;
}


void stream_cam(int me) {
  int other = me == 0 ? 1 : 0;

  printf("cam %d\n", me);

  pthread_mutex_lock(&mutex);
  running[other] = 0;
  running[me] = 1;
  pthread_cond_broadcast(&cond);
  pthread_mutex_unlock(&mutex);
}

void stop_cams() {
  pthread_mutex_lock(&mutex);
  running[0] = 0;
  running[1] = 0;
  pthread_cond_broadcast(&cond);
  pthread_mutex_unlock(&mutex);

  sleep(3);

  //system("ffmpeg -y -framerate 30 -i output_%04d.jpg vid0.mp4");
  deinit_mux();
}


long send_file_callback(void *cls, uint64_t pos, char *buf, size_t max) {
    FILE *fp = (FILE *)cls;
    size_t n = fread(buf, 1, max, fp);
    if (n == 0 && ferror(fp)) {
        return MHD_CONTENT_READER_END_WITH_ERROR;
    } else if (n == 0) {
        return MHD_CONTENT_READER_END_OF_STREAM;
    } else {
        return n;
    }
}

int serve_file(struct MHD_Connection *connection, const char *path) {
    FILE *fp = fopen(path, "rb");
    if (!fp) {
        return MHD_NO;
    }

    struct MHD_Response *response = MHD_create_response_from_callback(
        MHD_SIZE_UNKNOWN, // size of response, set to MHD_SIZE_UNKNOWN for unknown size
        4096,             // block size for callback function
        &send_file_callback,
        fp,               // file pointer passed to callback function
        (MHD_ContentReaderFreeCallback) &fclose);         // callback function to close the file pointer


    MHD_add_response_header(response, "Content-Type", "video/mp4");

    int ret = MHD_queue_response(connection, MHD_HTTP_OK, response);
    MHD_destroy_response(response);
    return ret;
}



static enum MHD_Result handle_request(void *cls, struct MHD_Connection *connection, const char *url, const char *method, const char *version, const char *upload_data, size_t *upload_data_size, void **con_cls) {
  if (strcmp(method, "GET") == 0 && strcmp(url, "/") == 0) {
    struct MHD_Response *response = MHD_create_response_from_buffer(strlen(page), (void *) page, MHD_RESPMEM_PERSISTENT);
    MHD_add_response_header(response, "Content-Type", "text/html");
    int ret = MHD_queue_response(connection, MHD_HTTP_OK, response);
    MHD_destroy_response(response);
    puts("got a request for the page");
    return ret;
  }
  else if (strcmp(method, "GET") == 0 && strcmp(url, "/cam1") == 0) {
    stream_cam(0);
    return empty_response(connection);
  }
  else if (strcmp(method, "GET") == 0 && strcmp(url, "/cam2") == 0) {
    stream_cam(1);
    return empty_response(connection);
  }
  else if (strcmp(method, "GET") == 0 && strcmp(url, "/stop") == 0) {
    puts("stop");
    stop_cams();
    sleep(1);
    static int nr_vids = 0;
    nr_vids++;
    char links[SZ_RESPONSE];
    create_video_part(nr_vids, links);
    
    struct MHD_Response *response = MHD_create_response_from_buffer(strlen(links), (void *) links, MHD_RESPMEM_MUST_COPY);
    MHD_add_response_header(response, "Content-Type", "text/html");
    int ret = MHD_queue_response(connection, MHD_HTTP_OK, response);
    MHD_destroy_response(response);
    return ret;
  }
  else if (strcmp(method, "GET") == 0 && strcmp(url, "/vid0.mp4") == 0) {
    puts("get the video");
    return serve_file(connection, "vid0.mp4");
  }
  else {
    //printf("else: %s %s %s\n", url, method, version);
    return MHD_NO;
  }
}



int main(int argc, char **argv) {
  pthread_t threads[NR_CAMS];
  
  int ids[NR_CAMS];

  for (int i = 0; i < NR_CAMS; i++) {
    ids[i] = i;
    pthread_create(&threads[i], NULL, handle_cam, &ids[i]);
  }

  int stream_server_fd = -1;

  int ret = start_stream_server(threads, &stream_server_fd);
  if (ret) {
    fprintf(stderr, "Failed to start webserver\n");
    return 1;
  }

  ret = init_mux("vid0.mp4");
  if (ret) {
    fprintf(stderr, "Failed to init muxer\n");
    return 1;
  }
    
  
  struct MHD_Daemon *daemon = MHD_start_daemon(MHD_USE_SELECT_INTERNALLY, 8080, NULL, NULL, &handle_request, NULL, MHD_OPTION_END);
  if (!daemon) {
    fprintf(stderr, "Failed to start webserver\n");
    return 1;
  }
  printf("Server running on port 8080...\n");
  getchar();
  MHD_stop_daemon(daemon);

  /* for (int i = 0; i < NR_CAMS; i++) { */
  /*   pthread_join(threads[i], NULL); */
  /* } */

  close(stream_server_fd);
  
  return 0;
}


